package it.unimi.di.mg4j.index;

/*		 
 * MG4J: Managing Gigabytes for Java
 *
 * Copyright (C) 2012 Sebastiano Vigna 
 *
 *  This library is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU Lesser General Public License as published by the Free
 *  Software Foundation; either version 3 of the License, or (at your option)
 *  any later version.
 *
 *  This library is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program; if not, see <http://www.gnu.org/licenses/>.
 *
 */


import static it.unimi.dsi.bits.Fast.INCR_STEP_8;
import static it.unimi.dsi.bits.Fast.MSBS_STEP_8;
import static it.unimi.dsi.bits.Fast.ONES_STEP_4;
import static it.unimi.dsi.bits.Fast.ONES_STEP_8;
import it.unimi.di.mg4j.index.payload.Payload;
import it.unimi.di.mg4j.search.IntervalIterator;
import it.unimi.di.mg4j.search.IntervalIterators;
import it.unimi.dsi.bits.Fast;
import it.unimi.dsi.fastutil.longs.LongBigList;
import it.unimi.dsi.fastutil.objects.Reference2ReferenceMap;
import it.unimi.dsi.fastutil.objects.Reference2ReferenceMaps;
import it.unimi.dsi.fastutil.objects.ReferenceSet;

import java.io.IOException;

/** An {@linkplain IndexReader index reader} for {@linkplain QuasiSuccinctIndex quasi-succinct indices}. */

public class QuasiSuccinctIndexReader extends AbstractIndexReader implements IndexReader {
	private final static boolean LAZY_COUNT_READING = true;

	/** An index iterator that can read Elias-Fano pointer lists. */
	protected final EliasFanoIndexIterator eliasFanoIndexIterator;
	/** An index iterator that can read ranked pointer lists. */
	protected final RankedIndexIterator rankedIndexIterator;
	/** The index, cached. */
	protected final QuasiSuccinctIndex index;
	/** The global current term ({@link #eliasFanoIndexIterator} and {@link #rankedIndexIterator} have both a similar local variable). */
	private int currentTerm;

	public QuasiSuccinctIndexReader( final QuasiSuccinctIndex index ) {
		this.index = index;
		eliasFanoIndexIterator = new EliasFanoIndexIterator( this );
		rankedIndexIterator = new RankedIndexIterator( this );
		currentTerm = -1;
	}
	
	protected final static class LongWordBitReader {

		private static final boolean DEBUG = false;

		/** The underlying list. */
		private final LongBigList list;
		/** The extraction width for {@link #extract()}. */
		private final int l;
		/** The extraction mask for {@link #l} bits. */
		private final long mask;

		/** The 64-bit buffer, whose lower {@link #filled} bits contain data. */
		private long buffer;
		/** The number of lower used bits {@link #buffer} (less than {@link Long#SIZE}). */
		private int filled;
		/** The current position in the list. */
		private long curr;

		public LongWordBitReader( final LongBigList list, final int l ) {
			this.list = list;
			this.l = l;
			mask = ( 1L << l ) - 1;
			curr = -1;
		}

		public LongWordBitReader position( final long position ) {
			if ( DEBUG ) System.err.println( this + ".position(" + position + ") [buffer = " +  Long.toBinaryString( buffer ) + ", filled = " + filled + "]" );
			final long wordPosition = position / Long.SIZE;
			final int bitPosition = (int)( position % Long.SIZE );

			// Quick (possibly empty) skip inside the current word.
			if ( wordPosition == curr && bitPosition >= Long.SIZE - filled ) {
				final long shift = bitPosition - Long.SIZE + filled;
				filled -= shift;
				buffer >>>= shift;
				return this;
			}
			curr = wordPosition;
			/* When the bit position is zero, we leave the buffer empty and we
			 * position on the previous longword (which could even be -1). The
			 * next access will increase curr and perform a read. */
			if ( bitPosition == 0 ) {
				buffer = filled = 0;
				curr--;
			}
			else {
				buffer = list.getLong( curr );
				buffer >>>= bitPosition;
				filled = Long.SIZE - bitPosition;
			}
			
			if ( DEBUG ) System.err.println( this + ".position() filled: " + filled + " buffer: " + Long.toBinaryString( buffer ));
			return this;
		}

		public long position() {
			return curr * Long.SIZE + Long.SIZE - filled; 
		}

		public long extract( final int width ) {
			if ( DEBUG ) System.err.println( this + ".extract(" + width + ") [buffer = " +  Long.toBinaryString( buffer ) + ", filled = " + filled + "]" );

			if ( width <= filled ) {
				long result = buffer & ( 1L << width ) - 1;
				filled -= width;
				buffer >>>= width;
				return result;
			}
			else {
				long result = buffer;
				buffer = list.getLong( ++curr );

				final int remainder = width - filled;
				if ( remainder == Long.SIZE ) {
					result = buffer;
					buffer = 0;
					return result;
				}
				result |= ( buffer & ( 1L << remainder ) - 1 ) << filled;
				buffer >>>= remainder;
				filled = Long.SIZE - remainder;
				return result;
			}
		}
		
		public long extract() {
			if ( DEBUG ) System.err.println( this + ".extract() " + l + " bits [buffer = " +  Long.toBinaryString( buffer ) + ", filled = " + filled + "]" );

			if ( l <= filled ) {
				long result = buffer & mask;
				filled -= l;
				buffer >>>= l;
				return result;
			}
			else {
				long result = buffer;
				buffer = list.getLong( ++curr );
				// Note that this WON'T WORK if remainder == Long.SIZE, but that's not going to happen.
				final int remainder = l - filled;
				result |= ( buffer & mask >>> filled ) << filled;
				buffer >>>= remainder;
				filled = Long.SIZE - remainder;
				return result;
			}
		}
		
		public int readUnary() {
			if ( DEBUG ) System.err.println( this + ".readUnary() [buffer = " +  Long.toBinaryString( buffer ) + ", filled = " + filled + "]" );

			int accumulated = 0;

			for(;;) {
				if ( buffer != 0 ) {
					final int msb = Long.numberOfTrailingZeros( buffer );
					filled -= msb + 1;
					/* msb + 1 can be Long.SIZE, so we must break down the shift. */
					buffer >>>= msb;
					buffer >>>= 1;
					if ( DEBUG ) System.err.println( this + ".readUnary() => " + ( msb + accumulated ) );
					return msb + accumulated;
				}
				accumulated += filled;
				buffer = list.getLong( ++curr );
				filled = Long.SIZE;
			}

		}

		public int readNegatedUnary() {
			if ( DEBUG ) System.err.println( this + ".readNegatedUnary() [buffer = " +  Long.toBinaryString( buffer ) + ", filled = " + filled + "]" );

			long word = ~( buffer | -1L << filled );

			int accumulated = 0;
			
			for(;;) {
				if ( word != 0 ) {
					final int msb = Long.numberOfTrailingZeros( word );
					filled -= msb + 1;
					/* msb + 1 can be Long.SIZE, so we must break down the shift. */
					buffer >>>= msb;
					buffer >>>= 1;
					if ( DEBUG ) System.err.println( this + ".readUnary() => " + ( msb + accumulated ) );
					return msb + accumulated;
				}
				accumulated += filled;
				word = ~( buffer = list.getLong( ++curr ) );
				filled = Long.SIZE;
			}

		}

		public int count( long toSkip ) {
			if ( DEBUG ) System.err.println( this + ".count(" + toSkip + ") [buffer = " +  Long.toBinaryString( buffer ) + ", filled = " + filled + "]" );
			int accumulated = 0;

			for(;;) {
				if ( toSkip < filled ) {
					accumulated += Long.bitCount( buffer & ( 1L << toSkip ) - 1 );
					buffer >>>= toSkip;
					filled -= toSkip;
					return accumulated;
				}
					
				toSkip -= filled;
				accumulated += Long.bitCount( buffer );
				buffer = list.getLong( ++curr );
				filled = Long.SIZE;
			}
		}
		
		public long readNonZeroGamma() {
			final int msb = readUnary();
			return extract( msb ) | ( 1L << msb );
		}

		public long readGamma() {
			return readNonZeroGamma() - 1;
		}

		public long readUnary( int k ) {
			if ( DEBUG ) System.err.println( this + ".readUnary(" + k + ") [buffer = " +  Long.toBinaryString( buffer ) + ", filled = " + filled + "]" );
			k--;
			for(;;) {
				long accumulator = 0;
				long byteSums;
				long word = buffer;

				for(;;) {
					// Phase 1: sums by byte
					byteSums = word - ( ( word & 0xa * ONES_STEP_4 ) >>> 1 );
					byteSums = ( byteSums & 3 * ONES_STEP_4 ) + ( ( byteSums >>> 2 ) & 3 * ONES_STEP_4 );
					byteSums = ( byteSums + ( byteSums >>> 4 ) ) & 0x0f * ONES_STEP_8;
					byteSums *= ONES_STEP_8;

					final int bitCount = (int)( byteSums >>> 56 );
					// System.err.println( "bitCount: " + bitCount + " byteSums: " + byteSums );
					if ( k < bitCount ) break;


					accumulator += filled - bitCount;
					word = list.getLong( ++curr );
					filled = Long.SIZE;
					k -= bitCount;
				}

				// System.err.println( "accumulator: " + accumulator + " byteSums: " + byteSums + " k: " + k  );

				// Phase 2: compare each byte sum with k
				final long residualStep8 = k * ONES_STEP_8;
				final long byteOffset = ( ( ( ( ( ( residualStep8 | MSBS_STEP_8 ) - ( byteSums & ~MSBS_STEP_8 ) ) ^ byteSums ^ residualStep8 ) & MSBS_STEP_8 ) >>> 7 ) * ONES_STEP_8 >>> 53 ) & ~0x7;

				// Phase 3: Locate the relevant byte and make 8 copies with incremental masks
				final int byteRank = (int)( k - ( ( ( byteSums << 8 ) >>> byteOffset ) & 0xFF ) );

				final long spreadBits = ( word >>> byteOffset & 0xFF ) * ONES_STEP_8 & INCR_STEP_8;
				final long bitSums = ( ( ( spreadBits | ( ( spreadBits | MSBS_STEP_8 ) - ONES_STEP_8 ) ) & MSBS_STEP_8 ) >>> 7 ) * ONES_STEP_8;

				// Compute the inside-byte location and return the sum
				final long byteRankStep8 = byteRank * ONES_STEP_8;

				final int select = (int)( byteOffset + ( ( ( ( ( byteRankStep8 | MSBS_STEP_8 ) - ( bitSums & ~MSBS_STEP_8 ) ^ bitSums ^ byteRankStep8 ) & MSBS_STEP_8 ) >>> 7 ) * ONES_STEP_8 >>> 56 ) );
				filled -= select + 1;

				/* select + 1 can be Long.SIZE, so we must break down the shift. */
				word >>>= select;
				buffer = word >>> 1;
				// System.err.println( "buffer: " + buffer + " filled: " + filled );

				return accumulator + select - k;
			}
		}

		public long readNegatedUnary( int k ) {
			if ( DEBUG ) System.err.println( this + ".readNegateUnary(" + k + ") [buffer = " +  Long.toBinaryString( buffer ) + ", filled = " + filled + "]" );
			long accumulator = 0;
			k--;
			for(;;) {
				long byteSums;
				long word = ~( buffer | -1L << filled );

				for(;;) {
					//System.err.println( "Z word: " + Long.toBinaryString( word ) + " " + filled );
					// Phase 1: sums by byte
					byteSums = word - ( ( word & 0xa * ONES_STEP_4 ) >>> 1 );
					byteSums = ( byteSums & 3 * ONES_STEP_4 ) + ( ( byteSums >>> 2 ) & 3 * ONES_STEP_4 );
					byteSums = ( byteSums + ( byteSums >>> 4 ) ) & 0x0f * ONES_STEP_8;
					byteSums *= ONES_STEP_8;

					final int bitCount = (int)( byteSums >>> 56 );
					//System.err.println( "Z bitCount: " + bitCount + " byteSums: " + byteSums );
					if ( k < bitCount ) break;

					accumulator += filled - bitCount;
					word = ~( buffer = list.getLong( ++curr ) );
					filled = Long.SIZE;
					k -= bitCount;
				}

				//System.err.println( "Z accumulator: " + accumulator + " byteSums: " + byteSums + " k: " + k  );

				// Phase 2: compare each byte sum with k
				final long residualStep8 = k * ONES_STEP_8;
				final long byteOffset = ( ( ( ( ( ( residualStep8 | MSBS_STEP_8 ) - ( byteSums & ~MSBS_STEP_8 ) ) ^ byteSums ^ residualStep8 ) & MSBS_STEP_8 ) >>> 7 ) * ONES_STEP_8 >>> 53 ) & ~0x7;

				// Phase 3: Locate the relevant byte and make 8 copies with incremental masks
				final int byteRank = (int)( k - ( ( ( byteSums << 8 ) >>> byteOffset ) & 0xFF ) );

				final long spreadBits = ( word >>> byteOffset & 0xFF ) * ONES_STEP_8 & INCR_STEP_8;
				final long bitSums = ( ( ( spreadBits | ( ( spreadBits | MSBS_STEP_8 ) - ONES_STEP_8 ) ) & MSBS_STEP_8 ) >>> 7 ) * ONES_STEP_8;

				// Compute the inside-byte location and return the sum
				final long byteRankStep8 = byteRank * ONES_STEP_8;

				final int select = (int)( byteOffset + ( ( ( ( ( byteRankStep8 | MSBS_STEP_8 ) - ( bitSums & ~MSBS_STEP_8 ) ^ bitSums ^ byteRankStep8 ) & MSBS_STEP_8 ) >>> 7 ) * ONES_STEP_8 >>> 56 ) );
				filled -= select + 1;
				/* select + 1 can be Long.SIZE, so we must break down the shift. */
				buffer >>>= select;
				buffer >>>= 1;
				//System.err.println( "Z accumulator: " + accumulator + " byteSums: " + byteSums + " k: " + k + " buffer : " + Long.toBinaryString( buffer ) + " filled: " + filled );
				return accumulator + select - k;
			}
		}
	}


	protected static class PointerReader {
		/** The longword bit reader for the upper bits. */
		protected final LongWordBitReader upperBits;
		/** The longword bit reader for pointers. */
		protected final LongWordBitReader skipPointers;
		/** The starting position of the pointers. */
		protected final long skipPointersStart;
		/** The starting position of the upper bits. */
		protected final long upperBitsStart;
		/** The logarithm of the quantum, cached from the index. */
		protected final int log2Quantum;
		/** The logarithm of the quantum, cached from the index. */
		protected final int quantum;
		/** The size of a pointer. */
		protected final int pointerSize;
		/** The number of pointers. */
		protected final long numberOfPointers;
		/** The frequency term (i.e., the number of elements of the current list). */
		protected final int frequency;
		/** The occurrency of the term. */
		protected final long occurrency;
		/** The index of the current prefix sum. */
		public long currentIndex;

		public PointerReader( final LongWordBitReader upperBits, final long upperBitsStart, final LongWordBitReader skipPointers, final long skipPointersStart, final long numberOfPointers, final int pointerSize, final int frequency, final long occurrency, final int log2Quantum ) {
			this.upperBits = upperBits;
			this.upperBitsStart = upperBitsStart;
			this.skipPointers = skipPointers;
			this.skipPointersStart = skipPointersStart;
			this.pointerSize = pointerSize;
			this.numberOfPointers = numberOfPointers;
			this.log2Quantum = log2Quantum;
			this.quantum = 1 << log2Quantum;
			this.occurrency = occurrency;
			this.frequency = frequency;
		}
	}


	protected final static class RankedPointerReader extends PointerReader {
		private static final boolean ASSERTS = false;
		private static final boolean DEBUG = false;
		/** The mask for the upper bits (above {@link #log2Quantum}). */
		private final long quantumUpperBitsMask;
		/** The mask for the lower bits (below {@link #log2Quantum}). */
		private final int quantumLowerBitsMask;
		/** The skipping threshold. */
		private final int skippingThreshold;
		/** The current prefix sum. */
		private long prefixSum;

		public RankedPointerReader( final LongWordBitReader upperBits, final long upperBitsStart, final LongWordBitReader skipPointers, final long skipPointersStart, final long numberOfPointers, final int pointerSize, final int frequency, final long occurrency, final int log2Quantum ) {
			super( upperBits, upperBitsStart, skipPointers, skipPointersStart, numberOfPointers, pointerSize, frequency, occurrency, log2Quantum );
			prefixSum = -1;
			quantumUpperBitsMask = -1L << log2Quantum;
			quantumLowerBitsMask = (int)~quantumUpperBitsMask;
			skippingThreshold = quantum * 2;
		}

		public long getNextPrefixSum() {
			//System.err.println( "Reading " + readUnary + " at position " + currentIndex );
			currentIndex++;
			prefixSum += upperBits.readUnary();
			return ++prefixSum;
		}

		public long skipTo( long lowerBound ) {
			if ( DEBUG ) System.err.println( this + ".skipTo(" + lowerBound + ") [currentIndex = " + currentIndex + ", prefixSum = " + prefixSum + ", frequency = " + frequency + "]" );

			long toSkip;
			if ( ( toSkip = lowerBound - ( upperBits.position() - upperBitsStart ) ) < skippingThreshold ) {
				currentIndex += upperBits.count( toSkip );
				currentIndex++;
				prefixSum = lowerBound + upperBits.readUnary();

				if ( ASSERTS ) { 
					final long position = upperBits.position() - upperBitsStart;
					upperBits.position( upperBitsStart );
					long c = 0;
					for( long i = position; i-- != 0; ) c += upperBits.extract( 1 );
					assert c == currentIndex : c + " != " + currentIndex;
					assert position + upperBitsStart == upperBits.position() : ( position + upperBitsStart ) + " != " + upperBits.position();
				} 

				if ( DEBUG ) System.err.println( this + ".skipTo(" + lowerBound + ") => " + prefixSum + "[currentIndex = " + currentIndex + "]" );
				return prefixSum;
			}

			upperBits.position( upperBitsStart + lowerBound );
			prefixSum = lowerBound + upperBits.readUnary();

			final long offset = prefixSum + 1;
			final long pointerIndex = offset >>> log2Quantum;
			currentIndex = ( pointerIndex == 0 ? 0 : skipPointers.position( skipPointersStart + ( pointerIndex - 1 ) * pointerSize ).extract( pointerSize ) );
			upperBits.position( upperBitsStart + ( offset & quantumUpperBitsMask ) );
			currentIndex += upperBits.count( offset & quantumLowerBitsMask ); 
			
			if ( ASSERTS ) { 
				final long position = upperBits.position() - upperBitsStart;
				upperBits.position( upperBitsStart );
				long c = 0;
				for( long i = position; i-- != 0; ) c += upperBits.extract( 1 );
				assert c == currentIndex : c + " != " + currentIndex;
				assert position + upperBitsStart == upperBits.position() : ( position + upperBitsStart ) + " != " + upperBits.position();
			} 
			
			if ( DEBUG ) System.err.println( this + ".skipTo(" + lowerBound + ") => " + prefixSum + "[currentIndex = " + currentIndex + "]" );
			return prefixSum;
		}		
	}
	
	
	protected final static class EliasFanoPointerReader extends PointerReader {
		private static final boolean ASSERTS = false;
		private static final boolean DEBUG = false;
		private final static int LINEAR_SCAN_THRESHOLD_BEFORE_RANDOM_ACCESS = 8;
		private final static int LINEAR_SCAN_THRESHOLD_AFTER_RANDOM_ACCESS = 16;
		
		/** The number of lower bits. */
		private final int l;
		/** The longword bit reader for the lower bits. */
		protected final LongWordBitReader lowerBits;
		/** The starting position of the power bits. */
		protected final long lowerBitsStart;
		/** The upper part of the current prefix sum. */
		private long upperPart;
		/** The lower part of the current prefix sum. */
		private long lowerPart;
		
		
		public EliasFanoPointerReader( final LongWordBitReader upperBits, final long upperBitsStart, final LongWordBitReader lowerBits, 
				final long lowerBitsStart, final int l, final LongWordBitReader skipPointers, final long skipPointersStart, final long numberOfPointers, final int pointerSize, final int frequency, final long occurrency, final int log2Quantum ) {
			super( upperBits, upperBitsStart, skipPointers, skipPointersStart, numberOfPointers, pointerSize, frequency, occurrency, log2Quantum );
			this.lowerBits = lowerBits;
			this.lowerBitsStart = lowerBitsStart;
			this.l = l;
			currentIndex = 0;
			upperPart = 0;
			lowerPart = -1;
		}

		public long getNextPrefixSum() {
			currentIndex++;
			return ( upperPart += upperBits.readUnary() ) << l | ( lowerPart = lowerBits.extract() );
		}
		
		public long skipTo( final long lowerBound ) {
			if ( DEBUG ) System.err.println( this + ".skipTo(" + lowerBound + ") [currentIndex=" + currentIndex + "]" );
			final LongWordBitReader upperBits = this.upperBits;
			final LongWordBitReader lowerBits = this.lowerBits;
			final long zeroesToSkip = lowerBound >>> l; // This is the index of the zero we want to find.
			long skippedZeroes = upperBits.position() - upperBitsStart - currentIndex;

			if ( zeroesToSkip - skippedZeroes < LINEAR_SCAN_THRESHOLD_BEFORE_RANDOM_ACCESS )
				for(;;) {
					final long prefixSum = upperPart << l | lowerPart;
					if ( prefixSum >= lowerBound ) return prefixSum;
					currentIndex++;
					upperPart += upperBits.readUnary();
					lowerPart = lowerBits.extract();
				}

            //System.err.println( this + ".skipTo(" + lowerBound + "): skippedZeroes = " + skippedZeroes + ", zeroesToSkip = " + zeroesToSkip + ", currentIndex = " + currentIndex + ", l = " + l );
			if ( zeroesToSkip - skippedZeroes > quantum ) {
				final long block = zeroesToSkip >>> log2Quantum;
				if ( ASSERTS ) assert block > 0;
				if ( ASSERTS ) assert block <= numberOfPointers;
				final long blockZeroes = block << log2Quantum;
				final long skip = skipPointers.position( skipPointersStart + ( block - 1 ) * pointerSize ).extract( pointerSize );
				if ( ASSERTS ) assert skip != 0;
				upperBits.position( upperBitsStart + skip );
				currentIndex = skip - blockZeroes;
				skippedZeroes = blockZeroes;

				if ( zeroesToSkip - skippedZeroes < LINEAR_SCAN_THRESHOLD_AFTER_RANDOM_ACCESS ) {
					upperPart = skippedZeroes + upperBits.readUnary(); 
					lowerPart = lowerBits.position( lowerBitsStart + l * currentIndex++ ).extract();
					for(;;) {
						final long prefixSum = upperPart << l | lowerPart;
						if ( prefixSum >= lowerBound ) return prefixSum;
						currentIndex++;
						upperPart += upperBits.readUnary();
						lowerPart = lowerBits.extract();
					}
				}
			}

			currentIndex += upperBits.readNegatedUnary( (int)( zeroesToSkip - skippedZeroes ) );
			upperPart = zeroesToSkip + upperBits.readUnary();
			lowerPart = lowerBits.position( lowerBitsStart + l * ( currentIndex++ ) ).extract();

			for(;;) {
				final long prefixSum = upperPart << l | lowerPart;
				if ( prefixSum >= lowerBound ) return prefixSum;
				currentIndex++;
				upperPart += upperBits.readUnary();
				lowerPart = lowerBits.extract();
			}
		}
		
		public String toString() {
			return this.getClass().getSimpleName() + '@' + Integer.toHexString( System.identityHashCode( this ) );
		}
	}

	
	protected final static class CountReader {
		//private static final boolean ASSERTS = true;
		private static final boolean DEBUG = false;
		
		/** The longword bit reader for pointers. */
		private final LongWordBitReader skipPointers;
		/** The longword bit reader for the lower bits. */
		private final LongWordBitReader lowerBits;
		/** The longword bit reader for the upper bits. */
		private final LongWordBitReader upperBits;

		/** The starting position of the pointers. */
		private final long skipPointersStart;
		/** The starting position of the power bits. */
		private final long lowerBitsStart;
		/** The starting position of the upper bits. */
		private final long upperBitsStart;
		
		/** The number of lower bits. */
		private final int l;
		/** The size of a pointer. */
		private final int pointerSize;
		/** The number of pointers. */
		private final long numberOfPointers;
		/** The logarithm of the quantum, cached from the index. */
		private final int log2Quantum;
		/** The quantum. */
		private final int quantum;

		/** The current prefix sum (the sum of the first {@link #currentIndex} elements). */
		public long prefixSum;
		/** The previous prefix sum. */
		public long prevPrefixSum;
		/** The index of the current prefix sum. */
		public long currentIndex;

		public CountReader( final LongBigList list, final long position, final int frequency, final long occurrency, final int log2Quantum ) {
			this.l = QuasiSuccinctIndexWriter.lowerBits( frequency, occurrency, true );
			this.pointerSize = QuasiSuccinctIndexWriter.pointerSize( frequency, occurrency, true, false );
			numberOfPointers = QuasiSuccinctIndexWriter.numberOfPointers( frequency, -1, log2Quantum, true, false );

			skipPointers = new LongWordBitReader( list, 0 );
			lowerBits = new LongWordBitReader( list, l );
			upperBits = new LongWordBitReader( list, 0 );
			
			skipPointersStart = position;
			lowerBitsStart = skipPointersStart + pointerSize * numberOfPointers;
			lowerBits.position( lowerBitsStart );
			upperBitsStart = lowerBitsStart + (long)l * frequency;
			upperBits.position( upperBitsStart );
			currentIndex = prevPrefixSum = prefixSum = 0;

			this.log2Quantum = log2Quantum;
			quantum = 1 << log2Quantum;
		}

		public long getLong( final long index ) {
			if ( DEBUG ) System.err.println( this + ".getLong(" + index + ") [currentIndex = " + currentIndex + "]" );
			long delta = index - currentIndex;

			if ( delta == 0 ) {	// shortcut
				currentIndex++;
				prefixSum = ( upperBits.readUnary() + ( ( prevPrefixSum = prefixSum ) >>> l ) ) << l | lowerBits.extract();
				if ( DEBUG ) System.err.println( this + ".getLong(" + index + ") shortcut => " + ( prefixSum - prevPrefixSum + 1 ) + " [currentIndex = " + currentIndex + ", prevPrefixSum = " + prevPrefixSum + ", prefixSum = " + prefixSum + "]" );
				return prefixSum - prevPrefixSum + 1;
			}

			currentIndex = index + 1;
			long accumulator;

			if ( delta < quantum ) accumulator = prefixSum >>> l;
			else {
				final long block = index >>> log2Quantum;
				final long skip = skipPointers.position( skipPointersStart + ( block - 1 ) * pointerSize ).extract( pointerSize );
				upperBits.position( upperBitsStart + skip );
				final long blockOnes = block << log2Quantum;
				accumulator = skip - blockOnes;
				delta = index - blockOnes;
			}

			switch( (int)delta ) {
				case 0: break;
				case 1: accumulator += upperBits.readUnary(); break;
				default: accumulator += upperBits.readUnary( (int)delta );
			}
			
			prevPrefixSum = accumulator << l | lowerBits.position( lowerBitsStart + l * ( index - 1 ) ).extract();
			prefixSum = ( accumulator + upperBits.readUnary() ) << l | lowerBits.extract();
			return prefixSum - prevPrefixSum + 1;
		}

		public String toString() {
			return this.getClass().getSimpleName() + '@' + Integer.toHexString( System.identityHashCode( this ) );
		}
	}

	protected final static class PositionReader {
		private static final boolean DEBUG = false;
		private static final int LINEAR_SCAN_THRESHOLD = 4;

		/** The longword bit reader for pointers. */
		private final LongWordBitReader skipPointers;
		/** The longword bit reader for the lower bits. */
		private final LongWordBitReader lowerBits;
		/** The longword bit reader for the upper bits. */
		private final LongWordBitReader upperBits;

		/** The starting position of the pointers. */
		private final long skipPointersStart;
		/** The starting position of the power bits. */
		private final long lowerBitsStart;
		/** The starting position of the upper bits. */
		private final long upperBitsStart;

		/** The number of lower bits. */
		private final int l;
		/** The size of a pointer. */
		private final int pointerSize;
		/** The number of pointers. */
		private final long numberOfPointers;
		/** The logarithm of the quantum, cached from the index. */
		private final int log2Quantum;
		/** The quantum. */
		private final int quantum;

		/** The current prefix sum (the sum of the first {@link #currentIndex} elements). */
		private long prefixSum;
		/** The index of the current prefix sum. */
		private long currentIndex;
		/** The base of the sequence of positions currently returned. */
		private long offset;

		public PositionReader( final LongBigList list, final long position, final long occurrency, final int log2Quantum ) {
			upperBits = new LongWordBitReader( list, 0 ).position( position );
			l = (int)upperBits.readGamma();
			numberOfPointers = QuasiSuccinctIndexWriter.numberOfPointers( occurrency, -1, log2Quantum, true, false );
			pointerSize = numberOfPointers == 0 ? -1 : (int)upperBits.readNonZeroGamma();

			skipPointers = new LongWordBitReader( list, pointerSize );
			lowerBits = new LongWordBitReader( list, l );;
			skipPointersStart = upperBits.position();
			lowerBitsStart = skipPointersStart + pointerSize * numberOfPointers;
			lowerBits.position( lowerBitsStart );
			upperBitsStart = lowerBitsStart + l * occurrency;
			upperBits.position( upperBitsStart );
			currentIndex = prefixSum = 0;

			this.log2Quantum = log2Quantum;
			quantum = 1 << log2Quantum;
		}

		public int getFirstPosition( long index ) {
			if ( DEBUG ) System.err.println( this + ".getFirstPosition(" + index + ")" );

			//System.err.println( "l: " + l  + " numpoint: " + numberOfPointers + " pointsize : " + pointerSize);
			long delta = index - currentIndex;
			currentIndex = index + 1;

			if ( delta < LINEAR_SCAN_THRESHOLD ) {
				while( delta-- != 0 ) prefixSum = ( upperBits.readUnary() + ( prefixSum  >>> l ) ) << l | lowerBits.extract();
				prefixSum = ( upperBits.readUnary() + ( ( offset = prefixSum ) >>> l ) ) << l | lowerBits.extract();
				return (int)( prefixSum - offset );
			}

			long accumulator;

			if ( delta < quantum ) accumulator = prefixSum >> l;
			else {
				final long block = index >>> log2Quantum;
				final long skip = skipPointers.position( skipPointersStart + ( block - 1 ) * pointerSize ).extract( pointerSize );
				upperBits.position( upperBitsStart + skip );
				final long blockOnes = block << log2Quantum;
				delta = index - blockOnes;
				accumulator = skip - blockOnes;
			}

			switch( (int)delta ) {
				case 0: break;
				case 1: accumulator += upperBits.readUnary(); break;
				default: accumulator += upperBits.readUnary( (int)delta );
			}

			offset = accumulator << l | lowerBits.position( lowerBitsStart + l * ( index - 1 ) ).extract();
			return (int)( ( prefixSum = ( accumulator + upperBits.readUnary() ) << l | lowerBits.extract() ) - offset );
		}
		
		public int getNextPosition() {
			currentIndex++;
			return (int)( ( prefixSum = ( ( upperBits.readUnary() + ( prefixSum >>> l ) ) << l | lowerBits.extract() ) ) - --offset );
		}
		
		public String toString() {
			return this.getClass().getSimpleName() + '@' + Integer.toHexString( System.identityHashCode( this ) );
		}
	}

	protected abstract static class AbstractQuasiSuccinctIndexIterator extends AbstractIndexIterator {
		private static final boolean ASSERTS = false;
		//private static final boolean DEBUG = false;
		/** The index reader associated to this index iterator. */
		protected final QuasiSuccinctIndexReader indexReader;
		/** The index of {@link #indexReader}, cached. */
		protected final QuasiSuccinctIndex index;
		/** A reference to the data for pointers. */
		protected final LongBigList pointersList;
		/** A reference to the data for counts, or <code>null</code> of {@link #hasCounts} is false. */
		protected final LongBigList countsList;
		/** A reference to the data for positions, or <code>null</code> of {@link #hasPositions} is false. */
		protected final LongBigList positionsList;
		/** A longword bit reader used for the upper bits of pointers and to check whether a posting list is ranked or not. */
		protected final LongWordBitReader pointersUpperBits;
		/** The count reader for the current term. */
		protected CountReader counts;
		/** The position reader for the current term. */
		protected PositionReader positions;
		/** An index interval iterator. */
		private final IndexIntervalIterator intervalIterator;
		/** A singleton set containing {@link #intervalIterator}. */
		private final Reference2ReferenceMap<Index,IntervalIterator> singletonIntervalIterator;
		/** The number of documents (cached from {@link #index}). */
		protected final int numberOfDocuments;
		/** The key index (cached from {@link #index}). */
		private final Index keyIndex;
		/** Cached from {@link #index}. */
		protected final boolean hasCounts;
		/** Cached from {@link #index}. */
		protected final boolean hasPositions;
		/** The current document. */
		protected int currentDocument;
		/** The frequency of the current term. */
		protected int frequency;
		/** The occurrency of the current term. */
		protected long occurrency;
		/** The index of the current term. */
		protected int currentTerm;
		/** The count of the current posting, or -1 if it is not known. */
		protected int count;
		/** The number of returned positions for the current document. */
		protected int nextPosition;

		protected AbstractQuasiSuccinctIndexIterator( final QuasiSuccinctIndexReader indexReader ) {
			this.indexReader = indexReader;
			index = indexReader.index;
			keyIndex = index.keyIndex;
			hasPositions = index.hasPositions;
			hasCounts = index.hasCounts;
			numberOfDocuments = index.numberOfDocuments;
			pointersList = index.getPointersList();
			countsList = hasCounts ? index.getCountsList() : null;
			positionsList = hasPositions ? index.getPositionsList() : null;
			pointersUpperBits = new LongWordBitReader( pointersList, 0 );

			intervalIterator = hasPositions ? new IndexIntervalIterator( this ) : null;
			singletonIntervalIterator = hasPositions ? Reference2ReferenceMaps.singleton( keyIndex, (IntervalIterator)intervalIterator ) : null;
			currentTerm = frequency = -1;
		}

		protected boolean ranked( final int term ) {
			//System.err.println( this + ".position(" + term + ")" );
			long position;

			if ( term == 0 ) position = 0;
			else  {
			    if ( index.pointersOffsets == null ) throw new IllegalStateException( "You cannot position an index without offsets" );
				position = index.pointersOffsets.getLong( term );
			}

			pointersUpperBits.position( position );
			occurrency = pointersUpperBits.readNonZeroGamma();
			frequency = occurrency == 1 ? 1 : (int)( occurrency - pointersUpperBits.readGamma() );
			int l = QuasiSuccinctIndexWriter.lowerBits( frequency + 1, numberOfDocuments, false );

			return frequency + 1L + ( numberOfDocuments >>> l ) + ( frequency + 1L ) * l > numberOfDocuments;
		}
		
		@Override
		public Index index() {
			return index;
		}

		@Override
		public int termNumber() {
			return currentTerm;
		}

		@Override
		public int frequency() throws IOException {
			return frequency;
		}

		@Override
		public Payload payload() throws IOException {
			throw new UnsupportedOperationException();
		}
		
		@Override
		public int nextPosition() throws IOException {
			if ( nextPosition == 0 ) {
				nextPosition = 1;
				count();
				return positions.getFirstPosition( counts.prevPrefixSum + counts.currentIndex - 1 );  
			}
			
			if ( nextPosition == count ) return END_OF_POSITIONS;
			nextPosition++;
			return positions.getNextPosition();
		}

		@Override
		public int document() {
			return currentDocument;
		}

		private void ensureHasPositions() {
			if ( ! hasPositions ) throw new UnsupportedOperationException( "Index " + index + " does not contain positions" );
		}

		private void ensureCurrentDocument() {
			if ( ( currentDocument | 0x80000000 ) == -1 ) throw new IllegalStateException( currentDocument == -1 ? "nextDocument() has never been called for (term=" + currentTerm + ")" : "This reader is positioned beyond the end of list of (term=" + currentTerm + ")" );
		}

		public Reference2ReferenceMap<Index,IntervalIterator> intervalIterators() throws IOException {
			intervalIterator();
			ensureHasPositions();
			return singletonIntervalIterator;
		}
		
		public IntervalIterator intervalIterator() throws IOException {
			if ( ! hasPositions ) return IntervalIterators.FALSE;
			if ( ASSERTS ) assert intervalIterator != null;
			return intervalIterator;
		}
		
		public IntervalIterator intervalIterator( final Index index ) throws IOException {
			ensureCurrentDocument();
			return index == keyIndex ? intervalIterator() : IntervalIterators.FALSE;
		}
		
		public ReferenceSet<Index> indices() {
			return index.singletonSet;
		}

		public String toString() {
			return index + " [" + currentTerm + "]" + ( weight != 1 ? "{" + weight + "}" : "" );
		}

		@Override
		public void dispose() throws IOException {
			indexReader.close();
		}
	}
	
	protected final static class EliasFanoIndexIterator extends AbstractQuasiSuccinctIndexIterator {
		private static final boolean DEBUG = false;
		/** The pointer reader for the current term. */
		protected EliasFanoPointerReader pointers;

		public EliasFanoIndexIterator( QuasiSuccinctIndexReader indexReader ) {
			super( indexReader );
		}

		protected boolean position( final int term ) {
			if ( ranked( term ) ) return false;
			
			final int log2Quantum = index.log2Quantum;
			final int l = QuasiSuccinctIndexWriter.lowerBits( frequency + 1, numberOfDocuments, false );
			final int pointerSize = QuasiSuccinctIndexWriter.pointerSize( frequency + 1, numberOfDocuments, false, true );
			final long numberOfPointers = QuasiSuccinctIndexWriter.numberOfPointers( frequency + 1, numberOfDocuments, log2Quantum, false, true );

			final LongWordBitReader skipPointers = new LongWordBitReader( pointersList, pointerSize );
			final LongWordBitReader lowerBits = new LongWordBitReader( pointersList, l );

			final long skipPointersStart = pointersUpperBits.position();
			final long lowerBitsStart = skipPointersStart + pointerSize * numberOfPointers;
			lowerBits.position( lowerBitsStart ); 
			final long upperBitsStart = lowerBitsStart + l * ( frequency + 1L ); // The +1 is for the terminator.
			pointersUpperBits.position( upperBitsStart );
						
			pointers = new EliasFanoPointerReader( pointersUpperBits, upperBitsStart, lowerBits, lowerBitsStart, l, skipPointers, skipPointersStart, numberOfPointers, pointerSize, frequency, occurrency, log2Quantum );

			if ( hasCounts ) {
				long position = term == 0 ? 0 : index.countsOffsets.getLong( term );
				counts = new CountReader( countsList, position, frequency, occurrency, log2Quantum );
				count = 0;
				
				if ( hasPositions ) {
					position = term == 0 ? 0 : index.positionsOffsets.getLong( term  );
					positions = new PositionReader( positionsList, position, occurrency, log2Quantum );  
				}
			}

			currentTerm = term;
			currentDocument = -1;
			return true;
		}
			
		@Override
		public int nextDocument() throws IOException {
			if ( currentDocument == END_OF_LIST ) return END_OF_LIST;
			if ( DEBUG ) System.err.println( this + ".nextDocument() [currentDocument = " + currentDocument + ", currentIndex = " + pointers.currentIndex + ", frequency = " + frequency + "]" );

			if ( LAZY_COUNT_READING ) {
				if ( ( currentDocument = (int)pointers.getNextPrefixSum() ) >= numberOfDocuments ) currentDocument = END_OF_LIST;
				count = nextPosition = 0;
			}
			else {
				if ( ( currentDocument = (int)pointers.getNextPrefixSum() ) >= numberOfDocuments ) currentDocument = END_OF_LIST;
				else {
					nextPosition = 0;
					count = (int)counts.getLong( pointers.currentIndex - 1 );
				}
			}
			if ( DEBUG ) System.err.println( this + ".nextDocument() => " + currentDocument );
			return currentDocument;
		}

		@Override
		public int skipTo( int n ) throws IOException {
			if ( n >= numberOfDocuments ) return currentDocument = END_OF_LIST;
			if ( DEBUG ) System.err.println( this + ".skipTo(" + n + ") [currentDocument = " + currentDocument + "]" );
			if ( currentDocument >= n ) return currentDocument;
			// We want to guarantee to return a document.
			if ( LAZY_COUNT_READING ) {
				if ( ( currentDocument = (int)pointers.skipTo( n ) ) >= numberOfDocuments ) currentDocument = END_OF_LIST;
				count = nextPosition = 0;
			}
			else {
				if ( ( currentDocument = (int)pointers.skipTo( n ) ) >= numberOfDocuments ) currentDocument = END_OF_LIST;
				else {
					nextPosition = 0;
					count = (int)counts.getLong( pointers.currentIndex - 1 );
				}
			}
			if ( DEBUG ) System.err.println( this + ".skipTo(" + n + ") => " + currentDocument );
			return currentDocument;
		}

		@Override
		public boolean mayHaveNext() {
			// System.err.println( pointers.currentIndex+" < "+frequency);
			return currentDocument != END_OF_LIST && pointers.currentIndex < frequency;
		}

		@Override
		public int count() throws IOException {
			//System.err.print( "Term " + termNumber()+ ", document " + currentDocument + ", posting " + (pointers.currentIndex - 1) + ", count: " + count + "..." );
			if ( LAZY_COUNT_READING ) if ( count == 0 ) count = (int)counts.getLong( pointers.currentIndex - 1 );
			//System.err.println( " => " + count );
			return count;
		}
	}
	
	protected final static class RankedIndexIterator extends AbstractQuasiSuccinctIndexIterator {
		private static final boolean DEBUG = false;
		/** The pointer reader for the current term. */
		protected RankedPointerReader pointers;

		public RankedIndexIterator( QuasiSuccinctIndexReader indexReader ) {
			super( indexReader );
		}

		protected boolean position( final int term ) {
			if ( ! ranked( term ) ) return false;
			
			final int log2Quantum = index.log2Quantum;
			
			final int pointerSize = Fast.length( frequency + 1 );
			final int numberOfPointers = numberOfDocuments >>> log2Quantum;

			final LongWordBitReader skipPointers = new LongWordBitReader( pointersList, pointerSize );
			final LongWordBitReader lowerBits = new LongWordBitReader( pointersList, 0 );

			final long skipPointersStart = pointersUpperBits.position();
			final long lowerBitsStart = skipPointersStart + pointerSize * numberOfPointers;
			lowerBits.position( lowerBitsStart ); 
			final long upperBitsStart = lowerBitsStart;
			pointersUpperBits.position( upperBitsStart );
						
			pointers = new RankedPointerReader( pointersUpperBits, upperBitsStart, skipPointers, skipPointersStart, numberOfPointers, pointerSize, frequency, occurrency, log2Quantum );

			if ( hasCounts ) {
				long position = term == 0 ? 0 : index.countsOffsets.getLong( term );
				counts = new CountReader( countsList, position, frequency, occurrency, log2Quantum );
				count = 0;
				
				if ( hasPositions ) {
					position = term == 0 ? 0 : index.positionsOffsets.getLong( term  );
					positions = new PositionReader( positionsList, position, occurrency, log2Quantum );  
				}
			}

			currentTerm = term;
			currentDocument = -1;
			return true;
		}
			
		@Override
		public int nextDocument() throws IOException {
			if ( currentDocument == END_OF_LIST ) return END_OF_LIST;
			if ( DEBUG ) System.err.println( this + ".nextDocument() [currentDocument = " + currentDocument + ", currentIndex = " + pointers.currentIndex + ", frequency = " + frequency + "]" );
			if ( LAZY_COUNT_READING ) {
				if ( ( currentDocument = (int)pointers.getNextPrefixSum() ) >= numberOfDocuments ) currentDocument = END_OF_LIST;
				count = nextPosition = 0;
			}
			else {
				if ( ( currentDocument = (int)pointers.getNextPrefixSum() ) >= numberOfDocuments ) currentDocument = END_OF_LIST;
				else {
					nextPosition = 0;
					count = (int)counts.getLong( pointers.currentIndex - 1 );
				}
			}
			if ( DEBUG ) System.err.println( this + ".nextDocument() => " + currentDocument );
			return currentDocument;
		}

		@Override
		public int skipTo( int n ) throws IOException {
			if ( n >= numberOfDocuments ) return currentDocument = END_OF_LIST;
			if ( DEBUG ) System.err.println( this + ".skipTo(" + n + ") [currentDocument = " + currentDocument + "]" );
			if ( currentDocument >= n ) return currentDocument;
			// We want to guarantee to return a document.
			if ( LAZY_COUNT_READING ) {
				if ( ( currentDocument = (int)pointers.skipTo( n ) ) >= numberOfDocuments ) currentDocument = END_OF_LIST;
				count = nextPosition = 0;
			}
			else {
				if ( ( currentDocument = (int)pointers.skipTo( n ) ) >= numberOfDocuments ) currentDocument = END_OF_LIST;
				else {
					nextPosition = 0;
					count = (int)counts.getLong( pointers.currentIndex - 1 );
				}
			}
			if ( DEBUG ) System.err.println( this + ".skipTo(" + n + ") => " + currentDocument );
			return currentDocument;
		}

		@Override
		public boolean mayHaveNext() {
			// System.err.println( pointers.currentIndex+" < "+frequency);
			return currentDocument != END_OF_LIST && pointers.currentIndex < frequency;
		}

		@Override
		public int count() throws IOException {
			//System.err.print( "Term " + termNumber()+ ", document " + currentDocument + ", posting " + (pointers.currentIndex - 1) + ", count: " + count + "..." );
			if ( LAZY_COUNT_READING ) if ( count == 0 ) count = (int)counts.getLong( pointers.currentIndex - 1 );
			//System.err.println( " => " + count );
			return count;
		}
	}

	private IndexIterator documents( final CharSequence term, final int termNumber ) {
		currentTerm = termNumber;
		if ( eliasFanoIndexIterator.position( termNumber ) ) {
			eliasFanoIndexIterator.term( term );
			return eliasFanoIndexIterator;
		}

		rankedIndexIterator.position( termNumber );
		rankedIndexIterator.term( term ); 
		return rankedIndexIterator;
	}

	@Override
	public IndexIterator documents( final int termNumber ) throws IOException {
		return documents( null, termNumber );
	}

	@Override
	public IndexIterator documents( final CharSequence term ) throws IOException {
		if ( closed ) throw new IllegalStateException( "This " + getClass().getSimpleName() + " has been closed" );
		if ( index.termMap != null ) {
			final int termIndex = (int)index.termMap.getLong( term );
			if ( termIndex == -1 ) return index.getEmptyIndexIterator( term, termIndex );
			return documents( term, termIndex );
		}
		throw new UnsupportedOperationException( "Index " + index + " has no term map" );
	}

	@Override
	public IndexIterator nextIterator() throws IOException {
		if ( currentTerm == index.numberOfTerms - 1 ) return null;
		return documents( ++currentTerm );
	}
	
}
